/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*-

  Copyright (C) 2002, 2005, 2006  Dmitry V. Levin <ldv@altlinux.org>
  Copyright (C) 2019-2021  Igor Vlasenko <viy@altlinux.org>

  Executes window manager using wm.d or xsessions database.

  This file is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.
*/

#ifndef _GNU_SOURCE
# define _GNU_SOURCE	1
#endif

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <error.h>
#include <unistd.h>
#include <getopt.h>

#include "wm-common.h"

static int list_mode;

static void
print_help()
{
	if (WM_LIST_DEBIAN_WM_SESSION == wm_list_mode)
		puts ("runwm is compiled to use /etc/X11/wmsession.d by default.");
	else
		puts ("runwm is compiled to use /usr/share/xsessions by default.");
	puts ("Usage: runwm <window manager> [args(ignored)]");
	puts ("       runwm --print <window manager> [args(ignored)]");
	puts ("       runwm --list");
	puts ("       runwm --help");
	puts ("\nExtra options:");
	puts (" --help	this help");
	puts (" -v,--verbose	verbose");
	puts (" -q,--quiet	quiet");
	puts (" --print	do nothing; just print command to execute");
	puts (" --alt,--altlinux | --deb,--debian	use /etc/X11/wmsession.d");
	puts (" --xdg,--freedesktop	use /usr/share/xsessions");
	puts ("");
}

int
main(int argc, char *argv[])
{
	/* commented; do not use getopt for better compatibility with original runwm
	int c;
	while (1)
	{
		static struct option long_options[] =
			{
				{"altlinux",    no_argument, &wm_list_mode, WM_LIST_DEBIAN_WM_SESSION},
				{"debian",      no_argument, &wm_list_mode, WM_LIST_DEBIAN_WM_SESSION},
				{"freedesktop", no_argument, &wm_list_mode, WM_LIST_XDG_WM_SESSION},
				{"xdg",         no_argument, &wm_list_mode, WM_LIST_XDG_WM_SESSION},
				{"print",       no_argument, &print_only, 1},
				{"list",       no_argument, &list_mode, 1},
				{"help",    no_argument,       0, 'h'},
				{"verbose", no_argument,       0, 'v'},
				{"quiet",   no_argument,       0, 'q'},
				{0, 0, 0, 0}
			};
		int option_index = 0;

		c = getopt_long (argc, argv, "hvq",
				 long_options, &option_index);

		if (c == -1)
			break;

		switch (c)
		{
		case 0:
			break;

		case 'h':
			print_help();
			exit(0);
			break;

		case 'v':
			verbose += 1;
			break;

		case 'q':
			verbose = 0;
			break;

		case '?':
			break;

		default:
			abort ();
		}
	}
	*/
	char *longopt;
	int optind = 1;
	while (argv[optind] && '-' == *argv[optind])
	{
		switch (*(argv[optind] + 1))
		{
		case '?':
		case 'h':
			print_help();
			exit(0);
			break;

		case 'v':
			verbose += 1;
			break;

		case 'q':
			verbose = 0;
			break;

		case '-':
			longopt = argv[optind] + 2;
			if (!strncmp("alt", longopt, 3) || !strncmp("deb", longopt, 3))
			{
				wm_list_mode = WM_LIST_DEBIAN_WM_SESSION;
			}
			else if (!strncmp("xdg", longopt, 3) || !strncmp("freedesktop", longopt, 11))
			{
				wm_list_mode = WM_LIST_XDG_WM_SESSION;
			}
			else if (!strncmp("help", longopt, 4))
			{
				print_help();
				exit(0);
			}
			else if (!strncmp("list", longopt, 4))
			{
				list_mode = 1;
			}
			else if (!strncmp("print", longopt, 5))
			{
				print_only = 1;
			}
			else
			{
				printf ("Invalid option: --%s", longopt);
				print_help();
				abort ();
			}
			break;

		default:
			printf ("Invalid option: -%c", *(argv[optind] + 1));
			print_help();
			abort ();
		}
		optind++;
	}

	if (optind >= argc && ! list_mode)
		error(EXIT_FAILURE, errno, "argument is required. See --help.");
	if (optind < argc && list_mode)
		error(EXIT_SUCCESS, errno, "--list mode: arguments are ignored.");

	ReadProgramConfig();

	wm_entry_list_header wm_header = PrepareWMEntryList();
	wm_entry  *entry_list = wm_header.entry_list;
	size_t i, total = GetWMEntryListTotal(&wm_header);

	if (list_mode)
	{
		for (i = 0; i < total; ++i)
			puts(entry_list[i]._runwmid);
		if (NULL == wm_entry_default)
			puts("default");
		return 0;
	}

	const char* wm=argv[optind];

	if (0 == strcmp(wm, "default"))
	{
		/* wm=default: use system default */
		if (NULL != wm_entry_default)
			ExecWMEntry(wm_entry_default);

		/* wm=default: use user saved */
		if (NULL != wm_entry_saved)
			ExecWMEntry(wm_entry_saved);

		/* wm=default: use user choice, not found in wm_list */
		ExecSaved();

		/* then prefdm: do we need it nowadays?
		choose WM by prefdm logic before:
		//prefdm_config=/etc/sysconfig/desktop
		if is_safe_config "$prefdm_config"; then
		...
		if (NULL != wm_entry_prefdm)
			ExecWMEntry(wm_entry_prefdm);
		*/

		/* select first; they are sorted by priority */
		if (0 != total)
			ExecWMEntry(&entry_list[0]);

		/* nothing to choose; try failsafe */
		if (wm_entry_failsafe)
			ExecWMEntry(wm_entry_failsafe);
	}

	if (0 == strcmp(wm, "failsafe") && wm_entry_failsafe)
		ExecWMEntry(wm_entry_failsafe);

	for (i = 0; i < total; ++i)
	{
		if (
			/* for non-XDG compliant DM; passes a string from runwm --list */
			0 == strcmp(wm, entry_list[i].wmname) ||
			/* for XDG compliant DM that pass <exec> from Exec=<exec> to Xsession
			 like LightDM, see https://bugzilla.altlinux.org/36913 */
			0 == strcmp(basename(wm), basename(entry_list[i].exec)) ||
			0 == strcmp(basename(wm), basename(entry_list[i].wmexec)))
		{
			ExecWMEntry(&entry_list[i]);
		}
	}

	error(EXIT_FAILURE, 0, "window manager \"%s\" not found.", wm);
	return EXIT_FAILURE;
}
