/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*-

  Copyright (C) 2002, 2005, 2006  Dmitry V. Levin <ldv@altlinux.org>
  Copyright (C) 2019-2021  Igor Vlasenko <viy@altlinux.org>

  The Window Manager's Selector common functions.

  This file is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.
*/

#ifndef _GNU_SOURCE
# define _GNU_SOURCE	1
#endif

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <errno.h>
#include <error.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <dirent.h>
#include <ctype.h>

#include "wm-common.h"

#define WM_ENTRY_XDG_DESKTOP 0
#define WM_ENTRY_DEBIAN_WM_SESSION 1

#define DEBIAN_WM_DIR "/etc/X11/wmsession.d"

#define PRIORITY_OF_DEFAULT_WM 10000;
#define PRIORITY_OF_SAVED_WM 2000000;

#ifdef WITH_WM_SESSION
int wm_list_mode = WM_LIST_DEBIAN_WM_SESSION;
#else /* ! WITH_WM_SESSION */
int wm_list_mode = WM_LIST_XDG_WM_SESSION;
#endif /* WITH_WM_SESSION */
int wm_xdg_use_wayland = 0;
int wm_xdg_use_x = 1;

int verbose = 1;

const char *saved = 0;
int print_only = 0;
int save_prefs = 0;

const wm_entry *wm_entry_default = NULL;
const wm_entry *wm_entry_selected = NULL;
const wm_entry *wm_entry_saved = NULL;
const wm_entry *wm_entry_failsafe = NULL;

/* failsafe entry palaceholder */
static
wm_entry failsafe_placeholder;

typedef struct
{
	char*  desktop_file;
	/* priority:
	   < 0 - strange/maybe invalid
	   0 - default
	   100-199 - not usable w/o RTFM or/and UTSL
	   usable w/o RTFM or/and UTSL:
	   400-999 - normal WM
	   400-499 - deprecated, historical
	   500-599 - contemporary non-xdg
	   600-699 - contemporary xdg
	   1000-9999 - (xdg) DE
	   1000-1999 - deprecated, historical, non-xdg (TDE)
	   2000-2999 - modern second league
	   3000-3999 - modern first league
	 */
	int  priority;
}
priority_entry;

static const char * const xdg_data_dirs_default[] = { "/usr/local/share", "/usr/share", NULL};
static const char * const bin_path_default[] = { "/bin", "/usr/local/bin", "/usr/bin", NULL};

#ifdef BUILTIN_PRIORITY
static const priority_entry xdg_priority_default[] = { {"plasma.desktop", 3900}, {"mate.desktop", 3800}, {NULL, 0} };
static const priority_entry* xdg_priority_db = xdg_priority_default;
#else
#include "wm_priority_generated.h"
static const priority_entry* xdg_priority_db = xdg_priority_generated;
#endif

static char *
getval(char *line, const char *name)
{
	size_t  namelen = strlen(name);

	if (!strncmp(line, name, namelen))
	{
		char   *val = line + namelen;
		size_t  len = strlen(val);

		if (len && ('\n' == val[len - 1]))
			val[--len] = '\0';
		return val;
	}

	return 0;
}

static char *
desktop_file_getval(char *line, const char *name)
{
	size_t  namelen = strlen(name);

	if (!strncmp(line, name, namelen))
	{
		char   *val = line + namelen;
		/* skip space before = sign */
		while(isspace(*val)) val++;
		/* = sign */
		if ('=' == *val)
			val++;
		else
			return 0;
		/* skip space after = sign */
		while(isspace(*val)) val++;
		size_t  len = strlen(val);

		if (len && ('\n' == val[len - 1]))
			val[--len] = '\0';
		return val;
	}

	return 0;
}

static char*
strdup_first_word_or_defval (const char* val, char* defval)
{
	const char *space_pos = NULL;
	const char *p;
	for (p = val; *p; p++)
		if (isspace(*p))
		{
			space_pos = p;
			break;
		}
	if (!space_pos)
		return defval;
	return strndup(val, space_pos - val);
}

static const char **
SplitByChar(char* buf, char separator)
{
	size_t s_count = 0, len = 0;
	const char *p = buf;
	while (*p)
	{
		len++;
		if (*p++ == separator)
			s_count++;
	}
	const char **list = calloc(s_count + 2, sizeof(char *));
	if (!list)
		error(EXIT_FAILURE, errno, "SplitByChar: alloc failed");
	list[0] = buf;
	size_t i, j = 0;
	for (i = 0; i<=len; i++)
	{
		if (separator == *(buf + i))
		{
			*(buf + i) = '\0';
			list[++j] = buf + i +1;
		}

	}
	list[s_count+1] = NULL;
	return list;
}

static char*
buffer_trim_space(char* buf)
{
	/* trim prefix space */
	while(isspace(*buf)) buf++;
	/* trim suffix space */
	char* atend = buf + strlen(buf) - 1;
	while(atend >= buf && isspace(*atend))
		*(atend--) = '\0';
	return buf;
}

static void
str_move_left(char *begin, char *end)
{
	char* p = begin;
	for (p = begin; p < end; p++)
		*p = *(p + 1);
}

static void
freedesktop_unescape_char(char** p, char** end)
{
	str_move_left(*p, *end);
	(*end)--;
	/* increments cancel each other: --(for str_move_left) and ++(for escape)
	  (*p)--;
	  (*p)++;
	*/
	if (**p == 'n')
		**p = '\n';
	else if (**p == 't')
		**p = '\t';
}

/*
  split into quoted strings with escape characters like in
  Exec=somerun --name "Soma\t-\t\"mommy\" warrior" --key \"mommy\"
   https://specifications.freedesktop.org/desktop-entry-spec/desktop-entry-spec-latest.html#exec-variables
*/
static const char **
freedesktop_split_string_for_exec_inplace(char *buf)
{
	/* trim prefix space */
	while(isspace(*buf)) buf++;

	size_t naive_word_count = 0;
	int in_space = 0;
	char *endbuf = buf;
	while (*endbuf)
	{
		if (isspace(*endbuf))
		{
			if (0 == in_space)
				naive_word_count++;
			in_space = 1;
		}
		else
		{
			in_space = 0;
		}
		endbuf++;
	}
	if (!in_space)
		naive_word_count++;

	/* +1 because words are terminated by NULL */
	const char **list = calloc(naive_word_count + 1, sizeof(char *));
	if (!list)
		error(EXIT_FAILURE, errno, "split_string_for_exec: alloc failed");

	const char* word = buf;
	size_t list_count = 0;
	char* p = buf;
	char *quoted_start = NULL;
	while (p < endbuf)
	{
		//error(EXIT_SUCCESS, 0, "p=%c in_space=%d quoted_start=%s", *p, in_space, quoted_start);
		if (isspace(*p))
		{
			if (! quoted_start)
			{
				if (0 == in_space)
				{
					*p = '\0';
					list[list_count++] = word;
				}
				in_space = 1;
			}
		}
		else
		{
			if (1 == in_space)
			{
				word = p;
			}
			in_space = 0;
			if (*p == '\\')
			{
				freedesktop_unescape_char(&p, &endbuf);
			}
			else if (*p == '"')
			{
				str_move_left(p, endbuf);
				p--;
				endbuf--;
				if (! quoted_start)
				{
					quoted_start = p + 1;
				}
				else
				{
					/* unescape(quoted_start, &p, &endbuf);*/
					char *q = quoted_start;
					while (q < p)
					{
						if (*q == '\\')
						{
							freedesktop_unescape_char(&q, &endbuf);
							p--;
						}
						q++;
					}
					quoted_start = NULL;
				}
			}
		}
		p++;
	}
	if (0 == in_space)
		list[list_count++] = word;
	list[list_count++] = NULL;
	return list;
}

static char *
str_concat2(const char *a, const char *b)
{
	char *both = malloc (strlen (a) + strlen (b) + 1);
	if (!both)
		error(EXIT_FAILURE, errno, "str_concat2 malloc");
	strcpy (both, a); strcat (both, b);
	return both;
}

static char *
xasprintf(char **ptr, const char *fmt, ...)
{
	va_list arg;

	va_start(arg, fmt);
	if (vasprintf(ptr, fmt, arg) < 0)
		error(EXIT_FAILURE, errno, "vasprintf");
	va_end(arg);

	return *ptr;
}

static wm_entry_list_header
NewWMEntryList()
{
	size_t initial_size = 32;
	wm_entry_list_header header = {0, 0, NULL};
	header.allocated = initial_size;
	if (!(header.entry_list = calloc(initial_size, sizeof(wm_entry))))
		error(EXIT_FAILURE, errno, "calloc");
	return header;
}

size_t
GetWMEntryListTotal(wm_entry_list_header *header)
{
	return header->total;
}

static void
PushWMEntryList(wm_entry_list_header *header, wm_entry* entry)
{
	if (header->allocated == header->total)
	{
		header->allocated *= 2;
		if (!(header->entry_list = reallocarray(header->entry_list, header->allocated, sizeof(wm_entry))))
			error(EXIT_FAILURE, errno, "realloc to %zd", header->allocated);
	}
	/*
	  https://specifications.freedesktop.org/basedir-spec/basedir-spec-latest.html

	  A specification that refers to $XDG_DATA_DIRS ... should define what the behaviour must be
	  when a file is located under multiple base directories. It could ... define that only the file
	  under the most important base directory should be used or ... rules for merging ...

	  we have XDG_BASE_DIRS (/usr/local/share:/usr/share) and we can't merge .desktop.
	  So first come first serve, other desktop files are rejected.
	*/
	wm_entry* list = header->entry_list;
	size_t i = 0;
	const size_t total = header->total;
	while (i < total)
		if (!strcmp(list[i++]._basename, entry->_basename))
		{
			if (VERBOSE_INFO)
				error(EXIT_SUCCESS, 0, "rejected %s: found %s", entry->_basename, list[i]._basename);
			return;
		}
	header->entry_list[total] = *entry;
	header->total++;
}

void
FreeWMEntry(wm_entry* entry)
{
	free(entry->file);
	if (entry->wmfile != entry->file)
		free(entry->wmfile);
	free(entry->name);
	if (entry->wmname != entry->name)
		free(entry->wmname);
	free(entry->type);
	free(entry->icon);
	free(entry->exec);
	if (entry->wmexec != entry->exec)
		free(entry->wmexec);
	free(entry->try_exec);
	free(entry->hidden);
	free(entry->xdg_current_session);
}

static int
SortWMEntry(const void* entry1, const void* entry2)
{
	if (((wm_entry*)entry1)->priority != ((wm_entry*)entry2)->priority)
	{
		if (((wm_entry*)entry1)->priority < ((wm_entry*)entry2)->priority)
			return 1;
		else
			return -1;
	}
	return strcmp(((wm_entry*)entry1)->wmfile, ((wm_entry*)entry2)->wmfile);
}

static int
GetDesktopBool(const char *val)
{
	if (!(val && *val))
		return 0;
	if (0 == strncmp("true",val,4))
		return 1;
	if (0 == strncmp("false",val,5))
		return 0;
	return atoi(val);
}

static void
ReadWMEntry(wm_entry *entry, char *fname, int entry_type, const priority_entry* priority_db)
{
	FILE *fp;
	char  buf[BUFSIZ];
	int   is_in_section = 0;
	char *nameprefix;
	char *iconprefix;
	char *execprefix;
	if (WM_ENTRY_DEBIAN_WM_SESSION == entry_type)
	{
		nameprefix = "NAME";
		iconprefix = "ICON";
		execprefix = "EXEC";
	}
	else
	{
		nameprefix = "Name";
		iconprefix = "Icon";
		execprefix = "Exec";
	}

	memset(entry, 0, sizeof(*entry));

	if (!fname)
		return;

	fp = fopen(fname, "r");
	if (!fp)
		return;

	entry->file = strdup(fname);
	entry->_basename = basename(entry->file);

	strncpy(buf, entry->_basename, sizeof(buf) - 1);
	size_t basename_len = strnlen(entry->_basename, sizeof buf);
	if (0 == strncmp(buf + basename_len - sizeof(".desktop") + 1, ".desktop", sizeof buf))
	{
		buf[basename_len - sizeof(".desktop") + 1] = '\0';
		/* trim XDG prefix: org.foo.bar -> bar */
		char *lastdot = strrchr(buf, '.');
		if (!lastdot)
			entry->wmfile = strdup(buf);
		else
			entry->wmfile = strdup(lastdot + 1);
		entry->priority = 0;
		while (priority_db && priority_db->desktop_file)
		{
			if (!strcmp(entry->_basename, priority_db->desktop_file))
			{
				entry->priority = priority_db->priority;
				break;
			}
			priority_db++;
		}
	}
	else if (WM_ENTRY_DEBIAN_WM_SESSION == entry_type &&
		 (buf[0] <= '9' && buf[0] >= '0' && buf[1] <= '9' && buf[1] >= '0'))
	{
		/* Debian/ALT wmsession.d file */
		entry->wmfile = strdup(buf + 2);
		buf[2] = 0;
		int debian_priority = atoi(buf);
		if (debian_priority)
			/* debian use the smaller the bigger priority */
			entry->priority = 100 - debian_priority;
		else
			entry->priority = 0;
	}
	else
	{
		entry->wmfile = strdup(entry->_basename);
		if (0 == strcmp("default", entry->_basename))
		{
			entry->priority = PRIORITY_OF_DEFAULT_WM;
		}
		else if (0 == strcmp("failsafe", entry->_basename))
			entry->priority = -1;
		else
		{
			entry->priority = -1;
			if (WM_ENTRY_DEBIAN_WM_SESSION == entry_type && VERBOSE_INFO)
				error(EXIT_SUCCESS, 0, "%s: bad wmsession.d file w/o priority", entry->file);
		}
	}

	while (fgets(buf, sizeof buf, fp))
	{
		char   *val;
		switch(buf[0])
		{
		case '[':
			if (WM_ENTRY_DEBIAN_WM_SESSION != entry_type)
			{
				if (is_in_section)
					/* is_in_section true; another section is started */
					goto close_entry_read;
				else
					if ((val = getval(buf, "[Desktop Entry]")))
					{
						is_in_section = 1;
						continue;
					}
			}
			break;
		case 'D':
			if ((val = desktop_file_getval(buf, "DesktopNames")))
			{
				val = buffer_trim_space(val);
				if (*val != '\0')
				{
					char *val_tail = val + strlen(val) - 1;
					/* trim suffix space and ; */
					while(val_tail >= val && (isspace(*val_tail) || ';' == *val_tail)) val_tail--;
					*(val_tail + 1) = '\0';
					for (char *val_i = val; val_i <= val_tail; val_i++)
					{
						if (*val_i == ';')
							*val_i = ':';
					}
					if (*val != '\0')
						entry->xdg_current_session = strdup(val);
				}
				continue;
			}
			break;
		case 'E':
			if ((val = desktop_file_getval(buf, execprefix)))
			{
				val = buffer_trim_space(val);
				if (*val != '\0')
				{
					entry->exec = strdup(val);
					entry->wmexec = strdup_first_word_or_defval(val, entry->exec);
				}
				continue;
			}
			break;
		case 'H':
			if ((val = desktop_file_getval(buf, "Hidden")))
			{
				val = buffer_trim_space(val);
				if (*val != '\0')
					entry->hidden = strdup(val);
				continue;
			}
			break;
		case 'I':
			if ((val = desktop_file_getval(buf, iconprefix)))
			{
				val = buffer_trim_space(val);
				if (*val != '\0' && *val != '/')
				{
					int len = strlen(val);
					if (0 == strcmp(val + len - 4, ".png") ||
					    0 == strcmp(val + len - 4, ".svg") ||
					    0 == strcmp(val + len - 4, ".xpm"))
					{
						if (VERBOSE_WARNING)
							error(EXIT_SUCCESS, 0,
							      "%s: bad Icon=%s: relative path with extension",
							      entry->file, val);
						*(val + len - 4) = '\0';
					}
				}
				if (*val != '\0')
					entry->icon = strdup(val);
				continue;
			}
			break;
		case 'N':
			if ((val = desktop_file_getval(buf, nameprefix)))
			{
				val = buffer_trim_space(val);
				/* icewm-session.desktop: 2 names, in [Desktop Entry] and [Window Manager] */
				if (!entry->name && (*val != '\0'))
				{
					entry->name = strdup(val);
					entry->wmname = strdup_first_word_or_defval(val, entry->name);
				}
				continue;
			}
			break;
		case 'S':
			if (WM_ENTRY_DEBIAN_WM_SESSION == entry_type && getval(buf, "SCRIPT:"))
			{
				goto close_entry_read;
			}
			break;
		case 'T':
			if ((val = desktop_file_getval(buf, "TryExec")))
			{
				val = buffer_trim_space(val);
				if (*val != '\0')
					entry->try_exec = strdup(val);
				continue;
			}
			if ((val = desktop_file_getval(buf, "Type")))
			{
				val = buffer_trim_space(val);
				if (*val != '\0')
					entry->type = strdup(val);
				continue;
			}
			break;
		}
	}
	close_entry_read:
	fclose(fp);

	if (0 == strcmp("default", entry->_basename) ||
	    0 == strcmp("default.desktop", entry->_basename))
	{
		entry->wmname = strdup("default"); /* wmname will be freed */
		entry->flags = entry->flags | WM_ENTRY_FLAG_DEFAULT;
		entry->priority = PRIORITY_OF_DEFAULT_WM;
	}
	else if (0 == strcmp("failsafe", entry->_basename) ||
	    0 == strcmp("failsafe.desktop", entry->_basename))
	{
		entry->wmname = strdup("failsafe"); /* wmname will be freed */
		entry->flags = entry->flags | WM_ENTRY_FLAG_FAILSAFE;
		entry->priority = -1;
	}
	if (saved && entry->exec && 0 == strcmp(saved, entry->exec))
	{
		entry->flags = entry->flags | WM_ENTRY_FLAG_SAVED;
		entry->priority = PRIORITY_OF_SAVED_WM;
	}

}

static int
CheckStat(const char* path)
{
	struct stat stb;
	int stat_retval = lstat(path, &stb);

	if (stat_retval < 0)
		return 0;

	if (!S_ISREG(stb.st_mode))
		return 0;

	return 1;
}

/* returns 1 if filename can be executed; return 0 otherwise */
/* PATH is simplified to /bin:/usr/bin */
static int
CheckIfExec(const char* filename)
{
	int can_exec = 1;
	if (*filename == '/')
		can_exec=access(filename, X_OK);
	else
	{
		const char * const* path = bin_path_default;
		do
		{
			char *try_exec_path_name = NULL;
			xasprintf(&try_exec_path_name, "%s/%s", *path, filename);
			can_exec=access(try_exec_path_name, X_OK);
			free(try_exec_path_name);
			if (can_exec >= 0)
				break;
		}
		while(*(++path));
	}
	if (can_exec < 0)
		return 0;
	return 1;
}

/* returns 1 if wm_entry is acceptable; return 0 otherwise */
static int
VerifyWMEntry(wm_entry* entry)
{
	if (!entry->file)
	{
		if (VERBOSE_INFO)
			error(EXIT_SUCCESS, 0, "%s: filesystem error: failed to read",
			      entry->file);
		return 0;
	}
	if (entry->hidden && GetDesktopBool(entry->hidden))
	{
		if (VERBOSE_INFO)
			error(EXIT_SUCCESS, 0, "%s: skipped (Hidden=%s)",
			      entry->file, entry->hidden);
		return 0;
	}
	if (entry->type && (
	    /* relaxed check (disabled) */
	    /* 0 == strcmp(entry->type, "Directory") || 0 == strcmp(entry->type, "Link") || */
	    /* strict check; allow XSession for some ugly .desktop examples */
		    (0 != strcmp(entry->type, "Application") && 0 != strcmp(entry->type, "XSession"))))
	{
		if (VERBOSE_INFO)
			error(EXIT_SUCCESS, 0, "%s: skipped (Type=%s)",
			      entry->file, entry->type);
		return 0;
	}
	if (entry->try_exec && ! CheckIfExec(entry->try_exec))
	{
		if (VERBOSE_INFO)
			error(EXIT_SUCCESS, 0, "%s: skipped (TryExec=%s)",
			      entry->file, entry->try_exec);
		return 0;
	}
	if (entry->wmexec && ! CheckIfExec(entry->wmexec))
	{
		if (verbose)
			error(EXIT_SUCCESS, 0, "%s: bad executable %s",
			      entry->file, entry->wmexec);
		return 0;
	}
	if (!entry->name || !entry->exec)
	{
		if (verbose)
			error(EXIT_SUCCESS, 0, "%s: incomplete config",
			      entry->file);
		return 0;
	}
	if (!entry->wmname || !entry->wmfile || !entry->wmexec || !entry->_basename)
	{
		if (verbose)
			error(EXIT_FAILURE, 0, "%s: internal failure",
			      entry->file);
		return 0;
	}
	return 1;
}

static wm_entry
NewWMEntry(const char* file, const char* name, const char* icon, const char* exec, int flags)
{
	wm_entry entry;
	entry.file = strdup(file);
	entry.wmfile = strdup(file);
	entry._basename = basename(entry.file);
	entry.name = strdup(name);
	entry.wmname = strdup_first_word_or_defval(name, entry.name);
	entry.type = NULL;
	entry.icon = NULL;
	if (icon)
		entry.icon = strdup(icon);
	entry.exec = strdup(exec);
	entry.wmexec = strdup_first_word_or_defval(exec, entry.exec);
	entry.try_exec = NULL;
	entry.hidden = NULL;
	entry.xdg_current_session = NULL;
	entry.priority = -1;
	entry.flags = flags;
	entry._runwmid = entry.name;
	//entry._wmlabel = entry.name;
	if (!VerifyWMEntry(&entry))
	    error(EXIT_FAILURE, 0, "NewWMEntry failed for %s: %s", entry.file, entry.name);
	return entry;
}

static void
ReadWMEntryDir(wm_entry_list_header* entry_list_header, const char* wm_dir, int(*CheckName)(const struct dirent *), void(*ReadWMEntry)(wm_entry*, char*))
{
	struct dirent *dir_ent;
	DIR * dirp = opendir(wm_dir);
	if (!dirp)
	{
		if (VERBOSE_INFO)
			error(EXIT_SUCCESS, 0, "can't open directory: %s", wm_dir);
		return;
	}

	while ((dir_ent = readdir(dirp)))
	{
		if (CheckName(dir_ent))
		{
			char   *fname;
			xasprintf(&fname, "%s/%s", wm_dir, dir_ent->d_name);
			if(CheckStat(fname))
			{
				wm_entry entry;
				ReadWMEntry(&entry, fname);
				if (VerifyWMEntry(&entry))
					PushWMEntryList(entry_list_header, &entry);
				else
					FreeWMEntry(&entry);
			}
			free(fname);
		}
	}
	closedir(dirp);
}

static int
CheckNameWMConfig(const struct dirent *ent)
{
	if (strchr(ent->d_name, '.'))
		return 0;
	return 1;
}

static void
ReadWMConfig(wm_entry *entry, char *fname)
{
	ReadWMEntry(entry, fname, WM_ENTRY_DEBIAN_WM_SESSION, NULL);
}

static wm_entry_list_header
ReadWMConfigDir()
{
	wm_entry_list_header header = NewWMEntryList();
	ReadWMEntryDir(&header, DEBIAN_WM_DIR, &CheckNameWMConfig, &ReadWMConfig);
	if (header.total <= 0)
		error(EXIT_SUCCESS, 0, "no sessions found in %s", DEBIAN_WM_DIR);
	return header;
}

static int
CheckNameXDGDesktop(const struct dirent *ent)
{
	int d_name_len = strlen(ent->d_name);
	/* 8 is strlen(".desktop"); 8+1(a char for a wm name)=9 */
	if (d_name_len < 9 || 0 != strcmp(ent->d_name + d_name_len - 8, ".desktop"))
		return 0;
	return 1;
}

static void
ReadWMXDGDesktop(wm_entry *entry, char *fname)
{
	ReadWMEntry(entry, fname, WM_ENTRY_XDG_DESKTOP, xdg_priority_db);
}

static wm_entry_list_header
ReadXDGWMDir()
{
	const char** xdg_data_dirs = (const char**) xdg_data_dirs_default;
	const char * xdg_data_dirs_env = getenv("XDG_DATA_DIRS");
	if (xdg_data_dirs_env && *xdg_data_dirs_env)
		xdg_data_dirs = SplitByChar(strdup(xdg_data_dirs_env), ':');
	if (xdg_data_dirs && VERBOSE_DEBUG)
	{
		const char * const* i = xdg_data_dirs;
		fprintf(stderr, "XDG_DATA_DIRS=%s", *i);
		while(*(++i))
			fprintf(stderr,":%s", *i);
		fputc('\n', stderr);
	}

	wm_entry_list_header header = NewWMEntryList();
	const char * const* i = xdg_data_dirs;
	do
	{
		char* wm_dir;
		if (wm_xdg_use_x)
		{
			wm_dir = str_concat2(*i, "/xsessions");
			ReadWMEntryDir(&header, wm_dir, &CheckNameXDGDesktop, &ReadWMXDGDesktop);
			free(wm_dir);
		}
		if (wm_xdg_use_wayland)
		{
			wm_dir = str_concat2(*i, "/wayland-sessions");
			ReadWMEntryDir(&header, wm_dir, &CheckNameXDGDesktop, &ReadWMXDGDesktop);
			free(wm_dir);
		}
	}
	while(*(++i));

	if (xdg_data_dirs != xdg_data_dirs_default)
		free(xdg_data_dirs);
	if (header.total <= 0)
		error(EXIT_SUCCESS, 0, "no XDG sessions found.");
	return header;
}

static wm_entry_list_header
CreateWMEntryList()
{
	if (WM_LIST_DEBIAN_WM_SESSION == wm_list_mode)
		return ReadWMConfigDir();
	return ReadXDGWMDir();
}

void
ExecWMEntry(const wm_entry *entry)
{
	if (!entry)
		error(EXIT_SUCCESS, 0, "ExecWMEntry: internal error: got nullptr");
	if (print_only)
	{
		puts(entry->exec);
		exit(0);
	}
	else
	{
		if (!print_only)
		{
			if (entry->xdg_current_session)
				setenv("XDG_CURRENT_DESKTOP",entry->xdg_current_session, 0);
			else
				setenv("XDG_CURRENT_DESKTOP","Old", 0);
		}
		if (VERBOSE_DEBUG)
			error(EXIT_SUCCESS, 0, "going execvp: %s", entry->exec);
		char* exec_buf = strdup(entry->exec);
		const char** exec_argv = freedesktop_split_string_for_exec_inplace(exec_buf);
		const char* exec_argc = exec_argv[0];
		exec_argv[0] = basename(exec_argc);
		if (VERBOSE_DEBUG)
			error(EXIT_SUCCESS, 0, "argc: %s", exec_argc);
		size_t i = 0;
		const char* p;
		while ((p = exec_argv[i]))
		{
			if (VERBOSE_DEBUG)
				error(EXIT_SUCCESS, 0, "argv[%zd]: %s", i, p);
			/* do freedesktop exec-variables substitutions here.
			   it doesn't have much sense for wm-select/runwm, but
			   1) wm-common can be reused into a general purpose library
			   2) some advanced freedesktop stuff can be introduced later
			   so keep this code here.
			*/
			if (i && '%' == *p && *(p+1) && !*(p+2))
			{
				switch(*(p+1))
				{
				case 'k':
					exec_argv[i] = entry->file;
					break;
				case 'c':
					/* TODO: re-implement with localized names support */
					exec_argv[i] = entry->name;
					break;
				/* just remove them */
				/* not implemented */
				case 'i':
				/* no arguments is passed from DM */
				case 'f':
				case 'F':
				case 'u':
				case 'U':
				/* deprecated */
				case 'd':
				case 'D':
				case 'n':
				case 'N':
				case 'v':
				case 'm':
					/* remove from the argv */
					if (VERBOSE_DEBUG)
						error(EXIT_SUCCESS, 0, "removed freedesktop field code %%%c", *(p+1));
					size_t j = i;
					while (exec_argv[j])
					{
						exec_argv[j] = exec_argv[j+1];
						j++;
					}
					break;
				}
			}
			i++;
		}
		if (VERBOSE_INFO)
		{
			i = 1;
			printf("executing: %s ", exec_argc);
			while (exec_argv[i])
			{
				printf("%s ", exec_argv[i++]);
			}
			putchar('\n');
		}
		execvp(exec_argc, (char* const*)exec_argv);
		error(EXIT_SUCCESS, errno, "execvp: %s", entry->exec);
		free(exec_buf);
	}
}

static char *
GetRCF()
{
	const char *home = getenv("HOME");
	char   *fname = 0;

	if (!home || !*home)
		return 0;

	xasprintf(&fname, "%s/.%s", home, program_invocation_short_name);

	return fname;
}

static void
SavePrefs(char* selected)
{
	char   *fname;
	FILE   *fp;

	if (!save_prefs)
		return;

	fname = GetRCF();
	if (!fname)
		return;

	fp = fopen(fname, "w");
	free(fname);
	if (!fp)
		return;

	fprintf(fp, "EXEC=%s\n", selected);
	fclose(fp);
}

void
ExecSelected()
{
	if (wm_entry_selected)
	{
		SavePrefs(wm_entry_selected->exec);
		ExecWMEntry(wm_entry_selected);
	}
}

void
ExecSaved()
{
	if (saved)
	{
		if (wm_entry_saved)
		{
			ExecWMEntry(wm_entry_saved);
		} else if (print_only)
		{
		    puts(saved);
		    exit(0);
		}
		else
		{
			execl(saved, basename(saved), NULL);
			error(EXIT_SUCCESS, errno, "execl: %s", saved);
		}
	}
	/* somehow Exec failed, saved is not reliable */
	saved = 0;
	wm_entry_saved = NULL;
}

void
ReadProgramConfig()
{
	char   *fname = GetRCF();
	FILE   *fp;
	char    buf[BUFSIZ];

	if (!fname)
		return;

	fp = fopen(fname, "r");
	free(fname);
	if (!fp)
		return;

	while (fgets(buf, sizeof buf, fp))
	{
		char   *val;

		if ((val = getval(buf, "EXEC=")))
		{
			saved = strdup(val);
			break;
		}
	}
	fclose(fp);
}

wm_entry_list_header
PrepareWMEntryList()
{
	wm_entry_list_header wm_header = CreateWMEntryList();
	wm_entry  *entry_list = wm_header.entry_list;
	size_t total = wm_header.total;
	qsort(entry_list, total, sizeof(wm_entry), &SortWMEntry);

	int* wmname_has_unicode = calloc(total, sizeof(int));
	int* wmname_is_not_unique = calloc(total, sizeof(int));
	int* wmfile_is_not_unique = calloc(total, sizeof(int));
	if (!wmname_is_not_unique || !wmfile_is_not_unique)
		error(EXIT_FAILURE, errno, "PrepareWMEntryList: malloc failed");

	size_t i, j;
	for (i = 0; i < total; ++i)
	{
		if (entry_list[i].flags & WM_ENTRY_FLAG_DEFAULT)
			wm_entry_default = &entry_list[i];
		if (entry_list[i].flags & WM_ENTRY_FLAG_FAILSAFE)
			wm_entry_failsafe = &entry_list[i];
		if (entry_list[i].flags & WM_ENTRY_FLAG_SAVED)
			wm_entry_saved = &entry_list[i];
	}
	if (!wm_entry_failsafe)
	{
		failsafe_placeholder = NewWMEntry(
		"failsafe.desktop",
		"failsafe",
		"failsafe",
		"xvt -fn fixed -geometry 80x24+0+0",
		WM_ENTRY_FLAG_REMOVE | WM_ENTRY_FLAG_FAILSAFE );
		wm_entry_failsafe = &failsafe_placeholder;
		if (VERBOSE_DEBUG)
			error(EXIT_SUCCESS, 0, "failsafe not found. created failsafe placeholder");

		/* if total == 0, failsafe becomes wisible */
		if (total == 0)
		{
			PushWMEntryList(&wm_header, &failsafe_placeholder);
			if (wm_header.total > 0)
				wm_entry_failsafe = &entry_list[0];
		}
	}

	for (i = 0; i < total; ++i)
	{
		if (entry_list[i].flags & WM_ENTRY_FLAG_DEFAULT)
			wm_entry_default = &entry_list[i];
		if (entry_list[i].flags & WM_ENTRY_FLAG_FAILSAFE)
			wm_entry_failsafe = &entry_list[i];
		if (entry_list[i].flags & WM_ENTRY_FLAG_SAVED)
			wm_entry_saved = &entry_list[i];
		/* check for unicode */
		char *p;
		for (p = entry_list[i].wmname; *p; p++)
		{
			if (!isalnum(*p) && !ispunct(*p))
			{
				wmname_has_unicode[i] = 1;
				break;
			}
		}
		/* check whether they are unique */
		for (j = i+1; j < total; ++j)
		{
			if (!strcmp(entry_list[i].wmname, entry_list[j].wmname))
			{
				wmname_is_not_unique[i] = 1;
				wmname_is_not_unique[j] = 1;
			}
			if (!strcmp(entry_list[i].wmfile, entry_list[j].wmfile))
			{
				wmfile_is_not_unique[i] = 1;
				wmfile_is_not_unique[j] = 1;
			}
		}
	}
	for (i = 0; i < total; ++i)
	{
		if (!entry_list[i]._runwmid)
		{
			if (!wmname_is_not_unique[i] && !wmname_has_unicode[i])
				entry_list[i]._runwmid = entry_list[i].wmname;
			else if (!wmfile_is_not_unique[i])
				entry_list[i]._runwmid = entry_list[i].wmfile;
			else
				entry_list[i]._runwmid = entry_list[i]._basename;
		}
	}

	return wm_header;
}
