/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*-

  Copyright (C) 2002, 2005, 2006  Dmitry V. Levin <ldv@altlinux.org>
  Copyright (C) 2019-2021  Igor Vlasenko <viy@altlinux.org>

  The Window Manager's Selector.

  This file is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.
*/

#ifndef _GNU_SOURCE
# define _GNU_SOURCE	1
#endif

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <error.h>
#include <locale.h>
#include <libintl.h>
#include <getopt.h>
#include <ctype.h>
#ifndef WITH_GTK3
  /* glib-2.0 deprecated GTypeDebugFlags/GTimeVal -> gtk2 doesn't compile with -Werror */
  /* or use pragma GCC diagnostic ignored "-Wdeprecated-declarations" */
  #define GTK_DISABLE_DEPRECATED
#endif
#include <gtk/gtk.h>
#include <gdk/gdkx.h>
/* from gqview: gqview-2.1.5/src/icons/sheet_broken.png, extended to 64x64 */
#include "image_notfound-32.xpm"
#include "image_notfound-48.xpm"
#include "image_notfound-64.xpm"
#include "wm-default-32.xpm"
#include "wm-default-48.xpm"
#include "wm-default-64.xpm"
#include "failsafe-32.xpm"
#include "failsafe-48.xpm"
#include "failsafe-64.xpm"

#define PACKAGE "wm-select"
#define LOCALEDIR "/usr/share/locale"
#define _(arg) gettext (arg)

#ifdef WITH_GTK3
  #define WM_SELECT_LABEL_FONT "label { font-family: Serif; font-size: 14px;}"
  #define WM_SELECT_WELCOME_FONT "label { font-family: Serif; font-size: 24px;}"
#else /* ! WITH_GTK3 */
  #define WM_SELECT_LABEL_FONT "Serif 14"
  #define WM_SELECT_WELCOME_FONT "Serif 24"
#endif /* WITH_GTK3 */
#define ICON_WIDTH 64

#include "wm-common.h"

static const char *
getLine(const char *fname, const char *def)
{
	char    buf[BUFSIZ];
	char   *line = 0;
	FILE   *fp = fopen(fname, "r");

	if (!fp)
		return def;

	if (fgets(buf, sizeof buf, fp))
	{
		size_t  len = strlen(buf);

		if (len && ('\n' == buf[len - 1]))
			buf[--len] = '\0';
		if (len)
			line = strdup(buf);
	}

	fclose(fp);
	return line ? : def;
}

static void
select_handler(GtkWidget *widget __attribute__ ((unused)), wm_entry *entry)
{
	wm_entry_selected = entry;
	gtk_main_quit();
}

static void
pref_handler(GtkWidget *widget __attribute__ ((unused)), GtkWidget * button)
{
	save_prefs = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(button));
}

static const char *
GetEnvString(const char *env_name, const char *default_string)
{
	const char *value = getenv(env_name);

	if (!value || !*value)
		value = default_string;

	return value;
}

#ifndef WITH_GTK3
static PangoFontDescription *
LoadFont(const char *font_spec)
{
	PangoFontDescription *font;

	font = pango_font_description_from_string(font_spec);
	if (!font)
		error(EXIT_FAILURE, errno ? : ENOENT, "LoadFont: %s", font_spec);

	return font;
}
#endif /* WITH_GTK3 */

static void
SetFont(GtkWidget * widget, const char *font_spec)
{
#ifdef WITH_GTK3
	/* see https://developer.gnome.org/gtk3/stable/TextWidget.html */
	GtkCssProvider *provider;
	GtkStyleContext *context;

	provider = gtk_css_provider_new();
	gtk_css_provider_load_from_data(provider, font_spec,
					-1,
					NULL);
	context = gtk_widget_get_style_context(widget);
	gtk_style_context_add_provider(context,
				       GTK_STYLE_PROVIDER (provider),
				       GTK_STYLE_PROVIDER_PRIORITY_APPLICATION);
#else /* ! WITH_GTK3 */
	GtkStyle *style = gtk_widget_get_style(widget);

	style->font_desc = LoadFont(font_spec);

	gtk_widget_set_style(widget, style);
#endif /* WITH_GTK3 */
}

static GdkPixbuf*
LookupIcon(char* entry_file, char* lookup_icon, const char ** failure_xpm, GtkIconTheme *icon_theme, int icon_size)
{
	GdkPixbuf *pixbuf = NULL;
	char* icon_names[] = { lookup_icon, NULL};
	/* or use GTK_ICON_LOOKUP_FORCE_SIZE flag here ? */
	GtkIconInfo* icon_info = gtk_icon_theme_choose_icon (icon_theme,(const char **) &icon_names, icon_size, 0);
	if (NULL != icon_info)
	{
		GError* gerror = NULL;
		pixbuf = gtk_icon_info_load_icon (icon_info, &gerror);
		if (gerror != NULL && verbose)
			error(EXIT_SUCCESS, 0, "%s: load icon from info failed for %s: %s",
			      basename(entry_file), lookup_icon, gerror->message);
		if (!pixbuf && failure_xpm)
			pixbuf = gdk_pixbuf_new_from_xpm_data (failure_xpm);
	}
	else if (failure_xpm)
	{
		pixbuf = gdk_pixbuf_new_from_xpm_data (failure_xpm);
		if (verbose)
			error(EXIT_SUCCESS, 0, "%s: Icon lookup failed for Icon=[%s]",
			      basename(entry_file), lookup_icon);
	}
	return pixbuf;
}

static GdkPixbuf*
GetPixbufForIcon(wm_entry *entry, GtkIconTheme *icon_theme, int icon_size)
{
	GdkPixbuf *pixbuf = NULL;
	char* session_icon = NULL;
	const char** wm_default_xpm;
	const char** failsafe_xpm;
	const char** image_notfound_xpm;
	switch (icon_size)
	{
	case 32:
		image_notfound_xpm = image_notfound_32_xpm;
		wm_default_xpm = wm_default_32_xpm;
		failsafe_xpm = failsafe_32_xpm;
		break;
	case 48:
		image_notfound_xpm = image_notfound_48_xpm;
		wm_default_xpm = wm_default_48_xpm;
		failsafe_xpm = failsafe_48_xpm;
		break;
	default:
	case 64:
		image_notfound_xpm = image_notfound_64_xpm;
		wm_default_xpm = wm_default_64_xpm;
		failsafe_xpm = failsafe_64_xpm;
		break;
	}
	if (entry->xdg_current_session)
	{
		char* p = session_icon = strdup(entry->xdg_current_session);
		for ( ; *p; ++p) {
			if (*p != ':')
				*p = tolower(*p);
			else
			{
				*p = '\0';
				break;
			}
		}
	}
	/* absolute path icon */
	if (entry->icon && *(entry->icon) == '/')
	{
		errno = 0;
		pixbuf = gdk_pixbuf_new_from_file_at_size(entry->icon, icon_size, icon_size, NULL);
		if (!pixbuf)
		{
			pixbuf = gdk_pixbuf_new_from_xpm_data (image_notfound_xpm);
			if(verbose)
			    error(EXIT_SUCCESS, errno, "%s: Error loading icon %s", basename(entry->file), entry->icon);
		}
	}
	else if (entry->icon)
	{
		if (!strcmp("failsafe", entry->icon))
			pixbuf = gdk_pixbuf_new_from_xpm_data (failsafe_xpm);
		else
			pixbuf = LookupIcon(entry->file, entry->icon, image_notfound_xpm, icon_theme, icon_size);
	}
	else if (session_icon)
	{
		pixbuf = LookupIcon(entry->file, session_icon, NULL, icon_theme, icon_size);
	}
	if (session_icon)
		free(session_icon);
	if (!pixbuf)
	{
		pixbuf = gdk_pixbuf_new_from_xpm_data (wm_default_xpm);
	}
	return pixbuf;
}

static GtkWidget *
MakeButton(wm_entry *entry, guint icon_width, guint button_padding)
{
	GtkWidget *vbox;
	GtkWidget *label;
	GtkWidget *pixmapwid;
	GtkWidget *button;
	GtkIconTheme *icon_theme = gtk_icon_theme_get_default();
#ifdef WITH_GTK3
	vbox = gtk_box_new(GTK_ORIENTATION_VERTICAL, 0);
#else /* ! WITH_GTK3 */
	vbox = gtk_vbox_new(FALSE, 0);
#endif /* WITH_GTK3 */
	gtk_container_set_border_width(GTK_CONTAINER(vbox), 1);

	pixmapwid = gtk_image_new_from_pixbuf(GetPixbufForIcon(entry, icon_theme, icon_width));
	gtk_box_pack_start(GTK_BOX(vbox), pixmapwid, 0, 0, button_padding);
	gtk_widget_show(pixmapwid);

	label = gtk_label_new(entry->wmname);
	gtk_box_pack_end(GTK_BOX(vbox), label, 0, 0, button_padding);
	gtk_widget_show(label);

	gtk_widget_show(vbox);

	button = gtk_button_new();
	g_signal_connect(button, "clicked",
			   G_CALLBACK(select_handler),
			   (gpointer) entry);
	gtk_container_add(GTK_CONTAINER(button), vbox);
	gtk_widget_show(button);

	return button;
}

static void
FillBox(GtkWidget * parent, wm_entry_list_header* wm_header, guint icon_width, guint button_padding)
{
#ifdef WITH_GTK3
	int grid_row = 0;
	GtkWidget *grid = gtk_grid_new();
#else /* ! WITH_GTK3 */
	GtkWidget *hbox = gtk_hbox_new(TRUE, 0);
#endif /* WITH_GTK3 */
	GtkWidget *button;

	wm_entry  *entry_list = wm_header->entry_list;
	int     i, total = GetWMEntryListTotal(wm_header), wrap_i = 0, wrap_limit;

	if (total <= 6)
		wrap_limit = total;
	else
	{
		if (total & 1)
			wrap_limit = (total + 1) / 2;
		else
			wrap_limit = total / 2;
	}
	// error (EXIT_SUCCESS, 0, "total=%d, wrap_limit=%d", total, wrap_limit);
	for (i = 0; i < total; ++i)
	{
		wm_entry *entry = &entry_list[i];

		button = MakeButton(entry, icon_width, button_padding);
#ifdef WITH_GTK3
		gtk_grid_attach(GTK_GRID(grid), button, wrap_i, grid_row, 1, 1);
#else /* ! WITH_GTK3 */
		gtk_box_pack_start(GTK_BOX(hbox), button, 0, 1, 0);
#endif /* WITH_GTK3 */
		gtk_widget_show(button);
		if (++wrap_i >= wrap_limit)
		{
#ifdef WITH_GTK3
			grid_row++;
#else /* ! WITH_GTK3 */
			gtk_box_pack_start(GTK_BOX(parent), hbox, 0, 0, 0);
			gtk_widget_show(hbox);
			hbox = gtk_hbox_new(TRUE, 0);
#endif /* WITH_GTK3 */
			wrap_i = 0;
		}
	}
#ifdef WITH_GTK3
	gtk_grid_set_row_homogeneous(GTK_GRID(grid), TRUE);
	gtk_grid_set_column_homogeneous(GTK_GRID(grid), TRUE);
	gtk_widget_show(grid);
	gtk_box_pack_start(GTK_BOX(parent), grid, 0, 0, 0);
#else /* ! WITH_GTK3 */
	if (wrap_i)
	{
		gtk_box_pack_start(GTK_BOX(parent), hbox, 0, 0, 0);
		gtk_widget_show(hbox);
	}
#endif /* WITH_GTK3 */

}

static GtkWidget *
MakeWelcomeLabel(GtkWidget * vbox, const char *text)
{
	GtkWidget *label = gtk_label_new(text);

	SetFont(label, GetEnvString("WM_SELECT_WELCOME_FONT", WM_SELECT_WELCOME_FONT));
#ifdef WITH_GTK3
	gtk_widget_set_halign(label, GTK_ALIGN_CENTER);
#else /* ! WITH_GTK3 */
	gtk_misc_set_alignment(GTK_MISC(label), 0.5, 0.5);
#endif /* WITH_GTK3 */
	gtk_box_pack_start(GTK_BOX(vbox), label, 0, 0, 10);
	gtk_widget_show(label);

	return label;
}

static void
MakeWindows(wm_entry_list_header* wm_header)
{
	GtkWidget *window;
	GtkWidget *button;
	GtkWidget *vbox;
	GtkWidget *hbox;
	GtkWidget *separator;

	guint button_padding = 5;
	guint icon_width = ICON_WIDTH;
#ifdef WITH_GTK3
#else /* ! WITH_GTK3 */
	gint screen_width  = gdk_screen_width();
	gint screen_height = gdk_screen_height();
	/* Raspberry Pi ? */
	if (screen_width <= 640 || screen_height <= 480)
		icon_width = 48;
	if (screen_width <= 320 || screen_height <= 240)
		icon_width = 32;
#endif /* WITH_GTK3 */
	switch (icon_width)
	{
	case 64:
		button_padding = 5;
		break;
	case 48:
		button_padding = 2;
		break;
	case 32:
		button_padding = 1;
		break;
	}
	window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_window_set_position(GTK_WINDOW(window), GTK_WIN_POS_CENTER_ALWAYS);
	gtk_container_set_border_width(GTK_CONTAINER(window), 10);
	gtk_window_set_title(GTK_WINDOW(window), "WM Selector");

	g_signal_connect(window, "destroy", G_CALLBACK(gtk_main_quit), window);
	g_signal_connect(window, "delete_event", G_CALLBACK(gtk_main_quit), window);

	gtk_widget_realize(window);

	SetFont(window,	GetEnvString("WM_SELECT_LABEL_FONT", WM_SELECT_LABEL_FONT));

#ifdef WITH_GTK3
	vbox = gtk_box_new(GTK_ORIENTATION_VERTICAL, 0);
#else /* ! WITH_GTK3 */
	vbox = gtk_vbox_new(FALSE, 0);
#endif /* WITH_GTK3 */

	MakeWelcomeLabel(vbox, _("Welcome to"));

	FillBox(vbox, wm_header, icon_width, button_padding);

	MakeWelcomeLabel(vbox, getLine("/etc/altlinux-release", "ALT Linux"));

#ifdef WITH_GTK3
	separator = gtk_separator_new(GTK_ORIENTATION_HORIZONTAL);
#else /* ! WITH_GTK3 */
	separator = gtk_hseparator_new();
#endif /* WITH_GTK3 */
	gtk_box_pack_start(GTK_BOX(vbox), separator, 0, 0, button_padding * 2);
	gtk_widget_show(separator);

#ifdef WITH_GTK3
	hbox = gtk_box_new(GTK_ORIENTATION_HORIZONTAL, 0);
#else /* ! WITH_GTK3 */
	hbox = gtk_hbox_new(FALSE, 0);
#endif /* WITH_GTK3 */

	button = gtk_check_button_new_with_label(_
						 ("Set default window manager"));
	g_signal_connect_object(button, "toggled",
				G_CALLBACK(pref_handler),
				button, 0);
	gtk_box_pack_start(GTK_BOX(hbox), button, 1, 1, 0);
	gtk_widget_show(button);

	button = gtk_button_new_with_label(_("Exit"));
	gtk_widget_set_can_default(button, TRUE);
	g_signal_connect_object(button, "clicked",
				G_CALLBACK(gtk_main_quit),
				window, 0);
	gtk_box_pack_end(GTK_BOX(hbox), button, 0, 0, 10);
	gtk_widget_show(button);

	gtk_box_pack_end(GTK_BOX(vbox), hbox, 0, 0, 0);
	gtk_widget_show(hbox);

	gtk_container_add(GTK_CONTAINER(window), vbox);
	gtk_widget_show(vbox);

	gtk_widget_show(window);
	gtk_widget_grab_default(button);
}

static void
InitLocale()
{
	const char *name = getenv("NLSPATH");

	if (!name || !*name)
		name = LOCALEDIR;

	if (!textdomain(PACKAGE))
		error(EXIT_FAILURE, errno, "textdomain");

	if (!bindtextdomain(PACKAGE, name))
		error(EXIT_FAILURE, errno, "bindtextdomain");

	if (!bind_textdomain_codeset(PACKAGE, "UTF-8"))
		error(EXIT_FAILURE, errno, "bind_textdomain_codeset");
}

int
main(int argc, char *argv[])
{
	/* from https://www.gnu.org/software/libc/manual/html_node/Getopt-Long-Option-Example.html */
	int c;
	while (1)
	{
		static struct option long_options[] =
			{
				/* These options set a flag. */
				{"altlinux",    no_argument, &wm_list_mode, WM_LIST_DEBIAN_WM_SESSION},
				{"debian",      no_argument, &wm_list_mode, WM_LIST_DEBIAN_WM_SESSION},
				{"freedesktop", no_argument, &wm_list_mode, WM_LIST_XDG_WM_SESSION},
				{"xdg",         no_argument, &wm_list_mode, WM_LIST_XDG_WM_SESSION},
				{"wayland",     no_argument, &wm_xdg_use_wayland, 1},
				{"no-wayland",  no_argument, &wm_xdg_use_wayland, 0},
				{"x",           no_argument, &wm_xdg_use_x, 1},
				{"no-x",        no_argument, &wm_xdg_use_x, 0},
				{"print",       no_argument, &print_only, 1},
				/* These options don’t set a flag.
				   We distinguish them by their indices. */
				{"help",    no_argument,       0, 'h'},
				{"verbose", no_argument,       0, 'v'},
				{"quiet",   no_argument,       0, 'q'},
				{0, 0, 0, 0}
			};
		/* getopt_long stores the option index here. */
		int option_index = 0;

		c = getopt_long (argc, argv, "hvq",
				 long_options, &option_index);

		/* Detect the end of the options. */
		if (c == -1)
			break;

		switch (c)
		{
		case 0:
			/* If this option set a flag, do nothing else now. */
			break;

		case 'h':
			puts ("Use:");
			puts ("wm-select [options]\n");
			puts ("Select and launch Window Manager/Desktop Environment for the X session.");
			puts ("wm-select supports both /etc/X11/wmsession.d and /usr/share/xsessions.");
			if (WM_LIST_DEBIAN_WM_SESSION == wm_list_mode)
				puts ("It is compiled to use /etc/X11/wmsession.d by default.");
			else
				puts ("It is compiled to use /usr/share/xsessions by default.");
			puts ("\nOptions:");
			puts (" --help	this help");
			puts (" -v,--verbose	verbose");
			puts (" -q,--quiet	quiet");
			puts (" --print	do nothing; just print command to execute");
			puts (" --alt,--altlinux | --deb,--debian	use /etc/X11/wmsession.d");
			puts (" --xdg,--freedesktop	use /usr/share/xsessions");
			puts (" --wayland	add wayland-sessions in xdg mode");
			puts (" --no-wayland	remove wayland-sessions in xdg mode");
			puts (" --x	add xsessions in xdg mode (default)");
			puts (" --no-x	remove xsessions in xdg mode");
			puts ("");
			exit(0);
			break;

		case 'v':
			verbose += 1;
			break;

		case 'q':
			verbose = 0;
			break;

		case '?':
			/* getopt_long already printed an error message. */
			break;

		default:
			abort ();
		}
	}

	InitLocale();

	ReadProgramConfig();
	wm_entry_list_header wm_header = PrepareWMEntryList();

	ExecSaved();

	gtk_init(&argc, &argv);

	MakeWindows(&wm_header);
	gtk_main();

	ExecSelected();
	return EXIT_FAILURE;
}
